"""Test web utils."""

from __future__ import annotations

import pytest


@pytest.mark.parametrize(
    "user_id, venue_id, brewery_id, beer_id, search, page, query, expected",
    [
        ("", "", "", "", False, "", None, "https://untappd.com"),
        ("123", "", "", "", False, "", None, "https://untappd.com/user/123"),
        ("", "456", "", "", False, "", None, "https://untappd.com/venue/456"),
        ("", "", "789", "", False, "", None, "https://untappd.com/brewery/789"),
        ("", "", "", "101112", False, "", None, "https://untappd.com/beer/101112"),
        ("", "", "", "", True, "", None, "https://untappd.com/search"),
        ("", "", "", "", False, "page1", None, "https://untappd.com/page1"),
        ("", "", "", "", False, "/v/page1", None, "https://untappd.com/v/page1"),
        ("", "", "", "", False, "", {"key": "value"}, "https://untappd.com?key=value"),
    ],
)
def test_url_of(
    user_id: str,
    venue_id: str,
    brewery_id: str,
    beer_id: str,
    search: bool,
    page: str,
    query: dict[str, str | int] | None,
    expected: str,
) -> None:
    """Test URL generation."""
    from untappd_scraper.web import url_of

    result = url_of(
        user_id=user_id,
        venue_id=venue_id,
        brewery_id=brewery_id,
        beer_id=beer_id,
        search=search,
        page=page,
        query=query,
    )

    assert result == expected
    assert result.startswith("https://untappd.com")



"""Test venue scraping."""

from __future__ import annotations

import pytest
from untappd_scraper.venue import Venue
from utpd_models_web.constants import UNTAPPD_VENUE_ACTIVITY_SIZE


@pytest.fixture
def venue_unv(_mock_venue_unv_get: None) -> Venue:
    """Eg, 4 Pines is unverified."""
    return Venue(14705)


@pytest.fixture
def venue_ver(_mock_venue_ver_get: None) -> Venue:
    """Eg, Sweeneys is verified with multuple menus."""
    return Venue(107565)


@pytest.fixture
def venue_nest(_mock_venue_ver_nest_get: None) -> Venue:
    """Eg, Dad & Daves have a menu pulldown."""
    return Venue(5840988)


# ----- Tests -----


def test_venue(venue_unv: Venue) -> None:
    result = venue_unv

    assert result
    assert result.name
    assert result.venue_id
    assert result.categories
    assert result.venue_name


@pytest.mark.usefixtures("_mock_venue_404")
def test_user_invalid() -> None:
    with pytest.raises(ValueError, match="Invalid"):
        Venue(123)  # ignored


def test_menus_unverified(venue_unv: Venue) -> None:
    result = venue_unv.menus()

    assert not result


@pytest.fixture
def venue(request: pytest.FixtureRequest) -> Venue:
    return request.getfixturevalue(request.param)


@pytest.mark.parametrize("venue", ["venue_ver", "venue_nest"], indirect=True)
def test_menus_verified(venue: Venue) -> None:
    menus = venue.menus()
    assert len(menus) >= 4

    result = next(iter(menus))

    assert result.beers
    assert result.full_name
    assert all("untappd" in r.beer_label_url for r in result.beers)
    assert all(r.brewery_id for r in result.beers)
    assert not any(r.brewery_url for r in result.beers)


def test_menus_named_verified(venue_ver: Venue) -> None:
    menus = venue_ver.menus("rooftop")
    assert len(menus) == 1

    result = next(iter(menus))

    assert result.name.casefold() == "rooftop"
    assert result.menu_id


@pytest.mark.xfail(reason="i think i need to patch /activity page")
def test_activity_verified(venue_ver: Venue) -> None:
    history = venue_ver.activity()
    assert history
    assert len(history) == UNTAPPD_VENUE_ACTIVITY_SIZE

    result = history[0]

    assert result.beer_id
    assert result.location
    assert result.beer_label_url
    assert result.brewery_url


def test_activity_unverified(venue_unv: Venue) -> None:
    history = venue_unv.activity()
    assert history
    assert len(history) == UNTAPPD_VENUE_ACTIVITY_SIZE

    result = history[0]

    assert result.beer_id
    assert result.location
    assert result.beer_label_url
    assert result.brewery_url


@pytest.mark.webtest
def test_menu_brewery_id() -> None:
    venue = Venue.from_name("sweeneys")
    assert venue
    menu = next(iter(venue.menus()))

    result = menu.beers

    assert result
    assert all(b.brewery_id for b in result)


@pytest.mark.webtest
def test_checkin_brewery_slug() -> None:
    venue = Venue.from_name("sweeneys")
    assert venue

    result = venue.activity()

    assert result
    assert all(b.brewery_url for b in result)



"""Untappd user beer history functions."""

from __future__ import annotations

import dataclasses
import logging
import random
import re
import time
from datetime import datetime
from typing import TYPE_CHECKING
from urllib.parse import urljoin

import parse
from pydantic import BaseModel, ConfigDict, computed_field
from utpd_models_web.beer import WebUserHistoryBeer
from utpd_models_web.constants import UNTAPPD_BASE_URL, UNTAPPD_BEER_HISTORY_SIZE

from untappd_scraper.logging_config import logger
from untappd_scraper.web import (
    date_from_details,
    end_of_href,
    make_soup,
    parse_abv,
    parse_ibu,
    parse_rating,
    url_of,
)

if TYPE_CHECKING:  # pragma: no cover
    from collections.abc import Iterable, Sequence

    from bs4 import BeautifulSoup, Tag


logging.getLogger("parse").setLevel(logging.INFO)

logger.info("Loading user beer history...")


def parse_web_user_history_beer(beer_item: Tag) -> WebUserHistoryBeer:
    """Parse beer details from a scraped history beer."""
    bid = int(str(beer_item["data-bid"]))

    if name_el := beer_item.select_one(".beer-details p.name a"):
        name = str(name_el.text).strip()
        url = urljoin(UNTAPPD_BASE_URL, str(name_el["href"]))
    else:  # pragma: no cover
        name = ""
        url = None

    label = str(img.get("src", "")) if (img := beer_item.select_one("a img")) else ""

    # Brewery
    brewery_el = beer_item.select_one(".beer-details p.brewery a")
    brewery_name = str(brewery_el.text.strip() if brewery_el else "")
    brewery_url = str(brewery_el["href"]).strip("/") if brewery_el else ""
    brewery_id = int(match[1]) if (match := re.search(r"/(\d+)$", brewery_url)) else None

    # Style
    style_el = beer_item.select_one(".beer-details p.style")
    style = str(style_el.text.strip() if style_el else "")

    # First/Recent check-in dates
    date_els = beer_item.select(".details p.date")
    if len(date_els) == 2:  # noqa: PLR2004
        first, recent = date_els

        first_checkin = date_from_details(first)
        recent_checkin = date_from_details(recent)
        first_checkin_id = int(end_of_href(first) or 0)
        recent_checkin_id = int(end_of_href(recent) or 0)
    else:  # pragma: no cover
        first_checkin = None
        first_checkin_id = None
        recent_checkin = None
        recent_checkin_id = None

    total_checkins = 0
    total_el = beer_item.select_one(".details p.check-ins")
    if total_el and (total := parse.search("Total: {:d}", str(total_el.text))):
        assert isinstance(total, parse.Result)
        total_checkins = total[0]

    # Ratings
    if ratings := beer_item.select_one(".ratings"):
        user_rating = parse_rating("their", ratings.text)
        global_rating = parse_rating("global", ratings.text)
    else:  # pragma: no cover
        user_rating = None
        global_rating = None

    # ABV
    abv_el = beer_item.select_one(".details p.abv")
    abv = parse_abv(str(abv_el.text)) if abv_el else None

    # IBU
    ibu_el = beer_item.select_one(".details p.ibu")
    ibu = parse_ibu(str(ibu_el.text)) if ibu_el else None

    return WebUserHistoryBeer(
        beer_id=bid,
        beer_name=name,
        beer_label_url=label,
        brewery_id=brewery_id,
        brewery_name=brewery_name,
        brewery_url=brewery_url,
        style=style,
        url=url,
        first_checkin=first_checkin or datetime.min,
        first_checkin_id=first_checkin_id or 0,
        recent_checkin=recent_checkin or datetime.min,
        recent_checkin_id=recent_checkin_id or 0,
        total_checkins=total_checkins,
        user_rating=user_rating,
        global_rating=global_rating,
        abv=float(abv) if abv else None,
        ibu=ibu,
    )


# Use Pyandtic to later interface with FastAPI and friends
class UserHistoryResponse(BaseModel):
    """Response for user beer history."""

    model_config = ConfigDict(frozen=True)

    total_expected: int
    results: list[WebUserHistoryBeer]

    @computed_field
    @property
    def total_found(self) -> int:
        """Return number of beers found."""
        return len(self.results)

    @computed_field
    @property
    def found_all(self) -> bool:
        """Return True if we found all beers."""
        return self.total_found >= self.total_expected

    def __repr__(self) -> str:
        """Don't show all the beers. There's heaps."""
        return (
            f"{self.__class__.__name__}("
            f"total_expected={self.total_expected}, "
            f"total_found={self.total_found}, "
            f"found_all={self.found_all}"
            ")"
        )


UserHistoryResponse.model_rebuild()


def beer_history(user_id: str) -> UserHistoryResponse:
    """Scrape the beer history of a user.

    This is expected to be the last 25 unique beers they've had.
    """
    soup, _ = make_soup(url_of(user_id, page="beers"))

    if stats := soup.select_one("div.stats a[data-href=':user/beerhistory'] span.stat"):
        total_uniques = int(stats.text.replace(",", ""))
    else:  # pragma: no cover
        total_uniques = UNTAPPD_BEER_HISTORY_SIZE

    return UserHistoryResponse(
        total_expected=total_uniques,
        results=[parse_web_user_history_beer(item) for item in soup.select(".beer-item")],
    )


def brewery_history(
    user_id: str, brewery_id: int, max_resorts: int = 0, *, switch_to_style_scrape: int = 3
) -> UserHistoryResponse:
    """Scrape the beer history of a user for a specific brewery.

    This will agressively re-sort the list (if requested) to extract more uniques.
    It will also filter by style once the number of unfound beers reduces.

    Args:
        user_id (str): user ID to load
        brewery_id (int): brewery ID to filter by
        max_resorts (int): how many times to re-sort the list to get more uniques
        switch_to_style_scrape (int): if this many missing styles, switch to style scrape

    Returns:
        UserHistoryResponse: all found beers user has had from this brewery
    """
    soup, _ = make_soup(url_of(user_id, page="beers", query={"brewery_id": brewery_id}))

    # add 1 to re-sorts to include initial sort "date"
    sort_keys = extract_sort_keys(soup, num_keys=max_resorts + 1, exclude=["brewery"])
    beers_per_brewery = calc_beers_per_brewery(soup)
    beers_per_style, style_to_id = calc_beers_per_style(soup)
    total_for_brewery = beers_per_brewery.get(brewery_id, 0)
    logger.debug("Max re-sorts: {}, sort_keys: {}", max_resorts, sort_keys)
    logger.debug("Beers per brewery {}: {}", brewery_id, total_for_brewery)

    # The first page already has unique beers. Processing below is for re-sorts
    beers = {parse_web_user_history_beer(item) for item in soup.select(".beer-item")}

    if (
        len(beers) >= UNTAPPD_BEER_HISTORY_SIZE  # filled a page, there may be more
        and len(beers) < total_for_brewery  # not all beers
        and max_resorts  # user requested re-sorts
    ):
        beers, missing_styles = run_brewery_resorts(
            ReSortContext(user_id, brewery_id, sort_keys, beers, switch_to_style_scrape),
            total_for_brewery=total_for_brewery,
            beers_per_style=beers_per_style,
        )

        if missing_styles:
            logger.debug("Switching to style scraping. Missing styles: {}", missing_styles)

            beers = run_style_resorts(
                ReSortContext(user_id, brewery_id, sort_keys, beers, switch_to_style_scrape),
                missing_styles=missing_styles,
                style_to_id=style_to_id,
            )

    return UserHistoryResponse(
        total_expected=total_for_brewery, results=sorted(beers, key=lambda x: x.beer_id)
    )


@dataclasses.dataclass(frozen=True)
class ReSortContext:
    """Args passed around for re-sorting."""

    user_id: str
    brewery_id: int
    sort_keys: list[str]
    beers: set[WebUserHistoryBeer]
    switch_to_style_scrape: int


def run_brewery_resorts(
    context: ReSortContext, *, total_for_brewery: int, beers_per_style: dict[str, int]
) -> tuple[set[WebUserHistoryBeer], dict[str, int]]:
    """Run the brewery re-sorts to get more beers."""
    missing_styles: dict[str, int] = {}

    for sort_key in context.sort_keys:
        if sort_key == "date":
            continue  # already tried this one implicitly
        logger.debug("Trying sort key {}", sort_key)

        soup, from_cache = make_soup(
            url_of(
                context.user_id,
                page="beers",
                query={"brewery_id": context.brewery_id, "sort": sort_key},
            )
        )
        fetched = {parse_web_user_history_beer(item) for item in soup.select(".beer-item")}
        context.beers.update(fetched)
        logger.debug("Now got {} beers", len(context.beers))

        if len(context.beers) >= total_for_brewery:
            logger.debug("Got all beers for brewery {}", context.brewery_id)
            break

        missing_styles = calc_missing_styles(context.beers, beers_per_style)
        if len(missing_styles) <= context.switch_to_style_scrape:
            logger.debug("Switching to style scraping. Missing styles: {}", missing_styles)
            break

        if from_cache:
            logger.debug("Page was cached, so no need to sleep")
        else:
            logger.debug("sleeping...")
            time.sleep(random.uniform(1, 3))  # noqa: S311
            logger.debug("awake")

    return context.beers, missing_styles


def run_style_resorts(
    context: ReSortContext, *, missing_styles: dict[str, int], style_to_id: dict[str, int]
) -> set[WebUserHistoryBeer]:
    """Run the style re-sorts to get more beers."""
    # Don't want to go wild here. Let's just get the "n" most popular styles
    styles = sorted(missing_styles.items(), key=lambda x: x[1], reverse=True)
    styles = styles[: context.switch_to_style_scrape]
    logger.debug("Trying styles: {}", styles)

    for style, num_in_style in styles:
        logger.debug("Trying style {} with {} beers", style, num_in_style)
        for sort_key in context.sort_keys:
            logger.debug("Trying sort key {} for style {}", sort_key, style)
            soup, from_cache = make_soup(
                url_of(
                    context.user_id,
                    page="beers",
                    query={
                        "brewery_id": context.brewery_id,
                        "sort": sort_key,
                        "type_id": style_to_id[style],
                    },
                )
            )
            fetched = {parse_web_user_history_beer(item) for item in soup.select(".beer-item")}
            context.beers.update(fetched)
            logger.debug("Now got {} beers", len(context.beers))

            beers_of_style = sum(b.style == style for b in context.beers)
            if beers_of_style >= num_in_style:
                logger.debug("Got all beers of style {}", style)
                break

            if from_cache:
                logger.debug("Page was cached, so no need to sleep")
            else:
                logger.debug("sleeping...")
                time.sleep(random.uniform(1, 3))  # noqa: S311
                logger.debug("awake")

    logger.debug("Done with style scraping")

    return context.beers


def extract_sort_keys(
    soup: BeautifulSoup, num_keys: int, *, exclude: Sequence[str] = ()
) -> list[str]:
    """Extract the sort keys from the page.

    Args:
        soup (BeautifulSoup): soup object to parse
        num_keys (NumReSorts): how many sort keys to return
        exclude (Sequence[str]): sort keys to exclude

    Returns:
        list[str]: list of sort keys to try, based on effort
    """
    sorting = soup.select_one("ul.menu-sorting")
    if not sorting:
        return []  # pragma: no cover
    sort_keys = [str(item["data-sort-key"]) for item in sorting.select("li.sort-items")]

    # "date" is default, so move it to the front
    if "date_asc" in sort_keys:
        sort_keys.remove("date_asc")
        sort_keys.insert(0, "date_asc")
    if "date" in sort_keys:
        sort_keys.remove("date")
        sort_keys.insert(0, "date")

    # remove any sort that matches our exclude list
    to_remove = [key for key in sort_keys if any(key.startswith(excl) for excl in exclude)]
    for key in to_remove:
        sort_keys.remove(key)

    return sort_keys[:num_keys]


def calc_beers_per_brewery(soup: BeautifulSoup) -> dict[int, int]:
    """Calculate the number of beers per brewery based of the pull down option text.

    Tag looks like `<option value="4792">Hallertau Brewery (3)</option>`

    Args:
        soup (BeautifulSoup): soup object to parse. Should be main beer history page.

    Returns:
        dict[int, int]: brewery ID to number of beers
    """
    beers: dict[int, int] = {}

    options = soup.select("#brewery_picker option")

    for option in options:
        if option["value"] == "all":
            continue
        brewery_id = int(str(option["value"]))
        if brewery_count := parse.search(r"({:d})", str(option.text)):
            assert isinstance(brewery_count, parse.Result)
            beers[brewery_id] = int(brewery_count[0])

    return beers


def calc_beers_per_style(soup: BeautifulSoup) -> tuple[dict[str, int], dict[str, int]]:
    """Calculate the number of beers per style based of the pull down option text.

    Tag looks like `<option value="128">IPA - American (12)</option>`

    Args:
        soup (BeautifulSoup): soup object to parse. Should be main beer history page.

    Returns:
        tuple[dict[str, int], dict[str, int]]: style to # beers, style to ID
    """
    beers: dict[str, int] = {}
    id_of_style: dict[str, int] = {}

    options = soup.select("#style_picker option")

    for option in options:
        if option["value"] == "all":
            continue
        style_info = parse.parse("{style} ({count:d})", str(option.text))
        if isinstance(style_info, parse.Result):
            beers[style_info["style"]] = int(style_info["count"])
            id_of_style[style_info["style"]] = int(str(option["value"]))

    return beers, id_of_style


def calc_missing_styles(
    beers: Iterable[WebUserHistoryBeer], beers_per_style: dict[str, int]
) -> dict[str, int]:
    """Calculate how many beers per style have not yet been scraped.

    Args:
        beers (Iterable[WebUserHistoryBeer]): beers already scraped
        beers_per_style (dict[str, int]): number of beers per style user has had
    """
    return {
        style: expected
        for style, expected in beers_per_style.items()
        if sum(b.style == style for b in beers) < expected
    }



"""Set up expectations for untappd-scraper API."""

from __future__ import annotations

from untappd_scraper.brewery import Brewery
from untappd_scraper.logging_config import configure_logging

logger = configure_logging(__name__)


brewery = Brewery.from_name("big niles")

print(brewery)  # noqa: T201



"""Untappd user functions."""

from __future__ import annotations

from dataclasses import dataclass, field, replace
from pathlib import PurePosixPath
from typing import TYPE_CHECKING, Any

from dateutil.parser import parse as parse_date
from httpx import HTTPStatusError
from utpd_models_web.checkin import WebActivityBeer
from utpd_models_web.user import WebUserDetails

from untappd_scraper.logging_config import configure_logging, logger
from untappd_scraper.user_beer_history import (
    UserHistoryResponse,
    beer_history,
    brewery_history,
)
from untappd_scraper.user_lists import WebUserList, load_user_lists, scrape_list_beers
from untappd_scraper.user_venue_history import load_user_venue_history
from untappd_scraper.web import get_text, make_soup, url_of

if TYPE_CHECKING:
    from collections.abc import Collection

    from bs4 import BeautifulSoup, Tag
    from utpd_models_web.venue import WebUserHistoryVenue

configure_logging(__name__)

logger.info("Loading user...")


@dataclass
class User:
    """Untappd user."""

    user_id: str

    details: WebUserDetails = field(init=False, repr=False)
    activity: tuple[WebActivityBeer, ...] = field(init=False, repr=False)

    # Only populate these when requested
    _lists: list[WebUserList] = field(default_factory=list, init=False, repr=False)
    _venue_history: list[WebUserHistoryVenue] = field(
        default_factory=list, init=False, repr=False
    )

    _soup_user: BeautifulSoup = field(init=False, repr=False)

    def __post_init__(self) -> None:
        """Initialise a User object, storing the user ID and loading details.

        Raises:
            ValueError: if the user ID is invalid or the page cannot be loaded
        """
        url = url_of(user_id=self.user_id)

        try:
            self._soup_user, _ = make_soup(url)
        except HTTPStatusError as e:
            logger.debug("Error getting user page for {}: {}", self.user_id, e)
            msg = f"Invalid userid {self.user_id}"
            raise ValueError(msg) from e

        self.activity = user_activity(self._soup_user)
        self.details = user_details(self._soup_user, url=url)

        # Populate recent activity summary into user details
        recent_beers = {a.beer_id for a in self.activity}
        recent_venues = {a.location_id for a in self.activity if a.location_id}
        self.details = replace(
            self.details, recent_beers=list(recent_beers), recent_venues=list(recent_venues)
        )

    @property
    def user_name(self) -> str:
        """Return the user's name."""
        return self.details.name

    def beer_history(self) -> UserHistoryResponse:
        """Scrape last 25 (or so) of a user's uniques.

        Returns:
            UserHistoryResponse: user's recent uniques
        """
        return beer_history(self.user_id)

    def brewery_history(self, brewery_id: int, max_resorts: int = 0) -> UserHistoryResponse:
        """Scrape as many of a user's uniques as possible from a brewery.

        Args:
            brewery_id (int): brewery id to filter by
            max_resorts (NumReSorts): number of times to re-sort the list to get more uniques

        Returns:
            UserHistoryResponse: user's uniques
        """
        return brewery_history(
            user_id=self.user_id, brewery_id=brewery_id, max_resorts=max_resorts
        )

    def lists(self) -> list[WebUserList]:
        """Scrape user's list page and return all visible listed beers.

        Returns:
            Collection[WebUserList]: all user's lists with 15 (or so) visible beers
        """
        if not self._lists:
            self._lists = load_user_lists(self.user_id)

        return self._lists

    def lists_detail(self, list_name: str) -> list[WebUserList]:
        """Return populated details of a user's list.

        Args:
            list_name (str): list name (or part thereof, case-insensitive)

        Returns:
            list[WebUserList]: matching lists with detail filled in
        """
        matching = [
            user_list
            for user_list in self.lists()
            if list_name.casefold() in user_list.name.casefold()
        ]

        for user_list in matching:
            user_list.beers.update(scrape_list_beers(user_list))

        return matching

    def venue_history(self) -> Collection[WebUserHistoryVenue]:
        """Scrape last 25 (or so) of a user's visited venues.

        Returns:
            Collection[WebUserHistoryVenue]: user's recent venues
        """
        if not self._venue_history:
            self._venue_history = load_user_venue_history(self.user_id)

        return self._venue_history

    def __getattr__(self, name: str) -> Any:  # noqa: ANN401
        """Return unknown attributes from user details.

        Args:
            name (str): attribute to lookup

        Returns:
            Any: attribute value
        """
        return getattr(self.details, name)


# ----- user details processing -----


def user_details(soup: BeautifulSoup, *, url: str) -> WebUserDetails:
    """Parse a user's main page into user details.

    Args:
        soup (BeautifulSoup): user's main page parsed
        url (str): URL of the user's page

    Returns:
        WebUserDetails: general user details
    """
    user_info = soup.select_one(".user-info .info")
    assert user_info, "User info tag not found"

    user_name = get_text(user_info, "h1")
    user_id = e.text.strip() if (e := user_info.select_one(".user-details p.username")) else ""
    user_location = (
        e.text.strip() if (e := user_info.select_one(".user-details p.location")) else ""
    )

    stats = soup.select_one(".stats")
    assert stats, "Stats tag not found"

    total_beers = (
        e.text if (e := stats.select_one('[data-href=":stats/general"] span')) else "0"
    )
    total_uniques = (
        e.text if (e := stats.select_one('[data-href=":stats/beerhistory"] span')) else "0"
    )
    total_badges = (
        e.text if (e := stats.select_one('[data-href=":stats/badges"] span')) else "0"
    )
    total_friends = (
        e.text if (e := stats.select_one('[data-href=":stats/friends"] span')) else "0"
    )

    return WebUserDetails(
        user_id=user_id,
        name=user_name,
        location=user_location,
        url=url,
        total_beers=str_to_int(total_beers),
        total_uniques=str_to_int(total_uniques),
        total_badges=str_to_int(total_badges),
        total_friends=str_to_int(total_friends),
    )


def user_activity(soup: BeautifulSoup) -> tuple[WebActivityBeer, ...]:
    """Parse a user's main page into checkins.

    Args:
        soup (BeautifulSoup): user's main page parsed

    Returns:
        tuple[WebActivityBeer, ...]: user's recent checkins
    """
    div = soup.select_one("div#main-stream")
    assert div, "Main stream tag not found"

    activities = div.select(".checkin")

    return tuple(parse_activity(activity) for activity in activities)


def parse_activity(activity: Tag) -> WebActivityBeer:
    """Parse a single activity item into a WebActivityBeer object."""
    checkin_link = activity.select_one("div.bottom a")
    assert checkin_link, "Checkin link tag not found"
    checkin_time = parse_date(checkin_link.text)
    assert checkin_time.tzinfo, "Checkin time does not have timezone info"

    top = activity.select_one("div.top")
    assert top, "Top tag not found"
    user_tag = top.select_one("p.text > a:nth-child(1)")
    assert user_tag, "User tag not found"
    beer_tag = top.select_one("p.text > a:nth-child(2)")
    assert beer_tag, "Beer tag not found"
    label_url = str(e["src"]) if (e := top.select_one("a img")) else ""
    brewery_tag = top.select_one("p.text > a:nth-child(3)")
    assert brewery_tag, "Brewery tag not found"
    location_tag = top.select_one("p.text > a:nth-child(4)")
    purchased_tag = top.select_one("p.purchased a")
    if rating_tag := top.select_one("div.caps[data-rating]"):
        data_rating = rating_tag["data-rating"]
        assert isinstance(data_rating, str), "Rating data is not a string"
        data_rating = float(data_rating)
    else:
        data_rating = None
    friends_tag = top.select_one("div.tagged-friends")
    friends = extract_friends(friends_tag)

    return WebActivityBeer(
        checkin_id=int(href_base(checkin_link)),
        checkin=checkin_time,
        user_name=href_base(user_tag),
        beer_name=beer_tag.text.strip(),
        beer_id=int(href_base(beer_tag)),
        beer_label_url=label_url,
        brewery_name=brewery_tag.text.strip(),
        brewery_id=None,  # not supplied in the activity
        brewery_url=href_base(brewery_tag),
        location=location_tag.text.strip() if location_tag else None,
        location_id=int(href_base(location_tag)) if location_tag else None,
        purchased_at=purchased_tag.text.strip() if purchased_tag else None,
        purchased_id=int(href_base(purchased_tag)) if purchased_tag else None,
        comment=e.text.strip() if (e := top.select_one("p.comment-text")) else None,
        serving=e.text.strip() if (e := top.select_one("p.serving")) else None,
        user_rating=data_rating,
        friends=friends,
    )


def str_to_int(numeric_string: str) -> int:
    """Convert a string to an integer.

    Args:
        numeric_string (str): amount with commas

    Returns:
        int: value as an integer
    """
    return int(numeric_string.replace(",", ""))


def href_base(link: Tag | None) -> str:
    """Extract the last bit from a link.

    Args:
        link (Tag): link element, with href="/a/b/123" for example

    Returns:
        str: ID extracted from the link
    """
    if not link:
        msg = "No link provided to href_base"
        logger.error(msg)
        raise ValueError(msg)

    href = link.get("href", "")
    assert isinstance(href, str), "Link href is not a string"

    return PurePosixPath(href).name


def extract_friends(friends_tag: Tag | None) -> list[str] | None:
    """Extract friends from a friends tag.

    Args:
        friends_tag (Tag | None): tag containing friends

    Returns:
        list[str] | None: list of friend names or None if no friends
    """
    if not friends_tag:
        return None

    return [href_base(friend) for friend in friends_tag.select("a") if href_base(friend)]



"""Untappd breweries."""

from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime
from typing import TYPE_CHECKING, Any, Self

from dateutil.parser import parse as parse_date
from httpx import HTTPStatusError
from utpd_models_web.brewery import WebBreweryDetails
from utpd_models_web.checkin import WebActivityBeer

from untappd_scraper.constants import UNTAPPD_BASE
from untappd_scraper.logging_config import configure_logging, logger
from untappd_scraper.web import (
    end_of_href,
    extract_checkin_elements_bs4,
    get_text,
    make_soup,
    parsed_value,
    url_of,
)

if TYPE_CHECKING:  # pragma: no cover
    from bs4 import BeautifulSoup


configure_logging(__name__)

logger.info("Loading brewery...")


@dataclass
class Brewery:
    """Untappd brewery."""

    brewery_id: int
    brewery_name: str = field(init=False)

    details: WebBreweryDetails = field(init=False, repr=False)
    activity: list[WebActivityBeer] = field(default_factory=list, init=False, repr=False)

    _soup: BeautifulSoup = field(init=False, repr=False)

    def __post_init__(self) -> None:
        """Initialise a Brewery object, storing the brewery ID and loading details.

        Raises:
            ValueError: invalid brewery ID
        """
        url = url_of(brewery_id=self.brewery_id)

        try:
            self._soup, _ = make_soup(url)
        except HTTPStatusError as e:
            logger.debug("Error getting brewery page for {}: {}", self.brewery_id, e)
            msg = f"Invalid brewery ID {self.brewery_id}"
            raise ValueError(msg) from e

        self.details = brewery_details(self._soup, brewery_id=self.brewery_id)
        self.brewery_name = self.details.name

        self.activity = brewery_activity(self._soup, brewery_id=self.brewery_id)

    def __getattr__(self, name: str) -> Any:  # noqa: ANN401
        """Return unknown attributes from venue details.

        Args:
            name (str): attribute to lookup

        Returns:
            Any: attribute value
        """
        return getattr(self.details, name)

    @classmethod
    def from_name(cls, brewery_name: str) -> Self | None:
        """Search for a brewery name and return the first one returned.

        Args:
            brewery_name (str): brewery name to search

        Returns:
            brewery: populated brewery object with first match of brewery name
        """
        url = url_of(search=True, query={"q": brewery_name, "type": "brewery"})
        soup, _ = make_soup(url)

        beers_link = soup.select_one("div.details.brewery a")

        if not beers_link:
            logger.debug("No brewery found for %s", brewery_name)
            return None

        brewery_id = parsed_value("/{:d}/beer", str(beers_link["href"]))

        return cls(int(brewery_id)) if brewery_id else None

    @classmethod
    def from_url(cls, brewery_url: str) -> Self | None:
        """Navigate via a brewery url and return the brewery objecct.

        Args:
            brewery_url (str): brewery url to load

        Returns:
            brewery: populated brewery object for this brewery url
        """
        url = url_of(page=brewery_url)
        try:
            soup, _ = make_soup(url)
        except HTTPStatusError as e:
            logger.debug("Error getting brewery page for url {}: {}", brewery_url, e)
            return None

        rss = soup.select_one("p.rss a")
        href = str(rss["href"]) if rss else ""

        if not href:  # pragma: no cover
            logger.debug("No brewery found for url {}", brewery_url)
            return None

        brewery_id = href.split("/")[-1]

        return cls(int(brewery_id))


# ----- brewery processing -----


def brewery_details(soup: BeautifulSoup, *, brewery_id: int) -> WebBreweryDetails:
    """Parse a brewery's main page into brewery details.

    Args:
        soup (BeautifulSoup): brewery's main page parsed
        brewery_id (int): brewery ID

    Returns:
        WebBreweryDetails: general brewery details
    """
    name = get_text(soup, "div.name h1")
    style = get_text(soup, "div.name p.style")
    address = get_text(soup, "div.name p.brewery")

    # Remove the "Show Less" text
    if desc_el := soup.select_one("div.desc div.beer-descrption-read-less"):
        for a in desc_el.select("a.read-less"):
            a.decompose()  # Remove from the tree
        description = desc_el.get_text(strip=True)
    else:
        description = ""  # pragma: no cover

    beers = get_text(soup, "div.details p.count")
    beers_parsed = parsed_value("{:d} beers", beers)
    num_beers = int(beers_parsed) if beers_parsed is not None else 0

    rating_tag = soup.select_one("div.details div.caps")
    rating = float(rating_tag.get("data-rating", 0)) if rating_tag else None  # pyright: ignore[reportArgumentType]

    if canonical := soup.select_one('link[rel="canonical"]'):
        url = str(canonical["href"]).rstrip("/")
        brewery_url = url.removeprefix(UNTAPPD_BASE)
    else:  # pragma: no cover
        brewery_url = f"brewery/{brewery_id}"

    return WebBreweryDetails(
        brewery_id=brewery_id,
        name=name,
        brewery_url=brewery_url,
        style=style,
        description=description,
        num_beers=num_beers,
        rating=rating,
        address=address,
    )


def brewery_activity(soup: BeautifulSoup, brewery_id: int) -> list[WebActivityBeer]:
    """Parse a brewery's activity page into checkins.

    Args:
        soup (BeautifulSoup): brewery's main page parsed
        brewery_id (int): brewery ID

    Returns:
        list[WebActivityBeer]: brewery activity
    """
    activity = []

    for beer in soup.select("div.activity div.item"):
        checkin_el = str(beer.get("data-checkin-id", ""))
        checkin_id = int(checkin_el) if checkin_el else 0
        time_el = beer.select_one(".time")
        checkin_time = parse_date(time_el.text) if time_el else datetime.min  # noqa: DTZ901
        label = img.get("src", "") if (img := soup.select_one("a.label img")) else ""

        user_el, beer_el, brewery_el, location_el = extract_checkin_elements_bs4(beer)
        purchased_el = beer.select_one("p.purchased a")

        if location_el and (location_href := end_of_href(location_el)):
            location_id = int(location_href)
        else:
            location_id = None
        if purchased_el and (purchased_href := end_of_href(purchased_el)):
            purchased_id = int(purchased_href)
        else:
            purchased_id = None

        purchased_at = str(purchased_el.text) if purchased_el else None

        rating_tag = soup.select_one("div.rating-serving div.caps")
        rating = float(rating_tag.get("data-rating", 0)) if rating_tag else None  # pyright: ignore[reportArgumentType]

        friends_tag = beer.select("div.tagged-friends a")
        friends = (
            [href for friend_tag in friends_tag if (href := end_of_href(friend_tag))]
            if friends_tag
            else None
        )

        brewery_url = (brewery_el.a and brewery_el.a.get("href")) or brewery_el.get("href")
        brewery_url = str(brewery_url).removesuffix("/") if brewery_url else ""

        activity.append(
            WebActivityBeer(
                checkin_id=checkin_id,
                checkin=checkin_time,
                user_name=end_of_href(user_el) or "",
                beer_name=beer_el.text,
                beer_label_url=str(label),
                beer_id=int(end_of_href(beer_el) or 0),
                brewery_name=brewery_el.text,
                brewery_id=brewery_id,
                brewery_url=brewery_url,
                location=location_el.text if location_el else None,
                location_id=location_id,
                purchased_at=purchased_at,
                purchased_id=purchased_id,
                comment=get_text(beer, "p.comment-text"),
                serving=get_text(beer, "p.serving"),
                user_rating=rating,
                friends=friends,
            )
        )

    return activity



"""Test user venue history scraping."""

from __future__ import annotations

import pytest
from untappd_scraper.user_venue_history import load_user_venue_history
from utpd_models_web.constants import UNTAPPD_VENUE_HISTORY_SIZE

# ----- Tests -----


@pytest.mark.usefixtures("_mock_user_venue_history_get")
def test_load_user_venue_history() -> None:
    venues = load_user_venue_history("test")
    assert len(venues) == UNTAPPD_VENUE_HISTORY_SIZE

    result = venues[0]

    assert result.name
    assert result.url


@pytest.mark.webtest
@pytest.mark.parametrize(
    "venue_id, expected",
    [
        (14705, False),  # 4 pines
        (107565, True),  # sweeney's
    ],
)
def test_recent_venue_verified(venue_id: int, expected: bool) -> None:
    venues = load_user_venue_history("mw1414")
    assert len(venues) == UNTAPPD_VENUE_HISTORY_SIZE

    result = next((v for v in venues if v.venue_id == venue_id), None)

    assert result
    assert result.is_verified == expected
